"""
Build WhatsApp Bots in Python • Fast, Effortless, Powerful 🚀

    - GitHub: https://github.com/david-lev/pywa
    - Docs: https://pywa.readthedocs.io/
    - PyPI: https://pypi.org/project/pywa/
"""

from pywa.client import WhatsApp
from pywa.utils import Version

__version__ = "3.0.0-rc.2"
__author__ = "David Lev"
__license__ = "MIT"
