"""
Build WhatsApp Bots in Python • Fast, Effortless, Powerful 🚀 [Async package]

    - GitHub: https://github.com/david-lev/pywa
    - Docs: https://pywa.readthedocs.io/
    - PyPI: https://pypi.org/project/pywa/
"""

from .client import WhatsApp
from .utils import Version
from pywa import __version__, __author__, __license__
